from pynwb.device import Device
from pynwb.icephys import IntracellularElectrode

from nwbinspector import Importance, InspectorMessage
from nwbinspector.checks import check_intracellular_electrode_cell_id_exists


def test_pass_check_intracellular_electrode_cell_id_exists():
    device = Device(name="device")
    ielec = IntracellularElectrode(name="ielec", cell_id="123", device=device, description="an intracellular electrode")
    assert check_intracellular_electrode_cell_id_exists(ielec) is None


def test_fail_check_intracellular_electrode_cell_id_exists():
    device = Device(name="device")
    ielec = IntracellularElectrode(name="ielec", device=device, description="an intracellular electrode")
    assert check_intracellular_electrode_cell_id_exists(ielec) == InspectorMessage(
        message="Please include a unique cell_id associated with this IntracellularElectrode.",
        importance=Importance.BEST_PRACTICE_VIOLATION,
        check_function_name="check_intracellular_electrode_cell_id_exists",
        object_type="IntracellularElectrode",
        object_name="ielec",
        location="/",
    )



import os
import sys
from pathlib import Path

from nwbinspector import available_checks
from collections import defaultdict

sys.path.append(str(Path(__file__).parent))
from conf_extlinks import extlinks, intersphinx_mapping

sys.path.insert(0, str(Path(__file__).resolve().parents[1]))

project = "NWBInspector"
copyright = "2022-2025, CatalystNeuro"
author = (
    "Cody Baker, Steph Prince, Szonja Weigl, Heberto Mayorquin, Paul Adkisson, Luiz Tauffer, Ryan Ly, and Ben Dichter"
)

extensions = [
    "sphinx.ext.doctest",
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx.ext.autosummary",
    "sphinx.ext.intersphinx",
    "sphinx.ext.extlinks",
    "sphinx_copybutton",
]
templates_path = ["_templates"]
master_doc = "index"
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]
html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]
html_css_files = ["custom.css"]

# These paths are either relative to html_static_path
# or fully qualified paths (eg. https://...)
html_css_files = [
    "css/custom.css",
]

html_theme_options = {
    "collapse_navigation": False,
}

# --------------------------------------------------
# Extension configuration
# --------------------------------------------------

# Napoleon
napoleon_google_docstring = False
napoleon_numpy_docstring = True
napoleon_use_param = False
napoleon_use_ivar = True
napoleon_include_init_with_doc = False
napoleon_include_private_with_doc = True
napoleon_include_special_with_doc = True

# Autodoc
autoclass_content = "both"
autodoc_member_order = "bysource"
autodata_content = "both"
autodoc_default_options = {
    "members": True,
    "member-order": "bysource",
    "private-members": True,
    "show-inheritance": False,
    "toctree": True,
}
add_module_names = False


def add_refs_to_docstrings(app, what, name, obj, options, lines):
    if what == "function" and obj.__name__.startswith("check_") and "Best Practice: " not in obj.__doc__:
        lines.append(f"Best Practice: :ref:`best_practice_{obj.__name__.split('check_')[1]}`")


def setup(app):
    app.connect("autodoc-process-docstring", add_refs_to_docstrings)


def _gen_checks_by_importance():
    dd = defaultdict(list)

    for x in available_checks:
        dd[x.importance.name.replace("_", " ")].append(x)

    generate_checks_rst_file = os.path.join(os.path.dirname(os.path.abspath(__file__)), "checks_by_importance.rst")
    with open(generate_checks_rst_file, "w") as f:
        f.write(
            """Checks by Importance
======================

This section lists the available checks organized by their importance level.

"""
        )

        for importance_level, checks in dd.items():
            f.write(
                f"""{importance_level}
{'-' * (len(f'{importance_level}') + 1)}

"""
            )

            for check in checks:
                f.write(f"*  :py:func:`~{check.__module__}.{check.__name__}`\n")

            f.write("\n")


_gen_checks_by_importance()



# Use this for mapping to external links
extlinks = {
    "pynwb-docs": ("https://pynwb.readthedocs.io/en/stable/%s", "%s"),
    "matnwb-src": ("https://github.com/NeurodataWithoutBorders/matnwb/%s", "%s"),
    "nwb-overview": ("https://nwb-overview.readthedocs.io/en/latest/%s", "%s"),
    "conda-install": (
        "https://docs.conda.io/projects/conda/en/latest/user-guide/install/index.html#regular-installation%s",
        "%s",
    ),
    "nwbinspector-issues": ("https://github.com/NeurodataWithoutBorders/nwbinspector/issues/%s", None),
    "nwbinspector-contributing": (
        "https://github.com/NeurodataWithoutBorders/nwbinspector/blob/dev/.github/CONTRIBUTING.md%s",
        None,
    ),
    "dandi-archive": ("https://dandiarchive.org/%s", "%s"),
    "ros3-tutorial": (
        "https://pynwb.readthedocs.io/en/stable/tutorials/advanced_io/streaming.html#streaming-method-2-ros3%s",
        None,
    ),
    "alternative-streaming-tutorial": (
        "https://pynwb.readthedocs.io/en/stable/tutorials/advanced_io/streaming.html#streaming-method-1-fsspec%s",
        None,
    ),
    "wikipedia": ("https://en.wikipedia.org/wiki/%s", "%s"),
    "allen-brain-map": ("https://%s.brain-map.org/", "%s"),
    "uuid": ("https://docs.python.org/3/library/uuid.html%s", None),
    "orcid": ("https://orcid.org/%s", "%s"),
    "pre-commit-bot": ("https://results.pre-commit.ci/%s", None),
    "neuroconv-coding-style": (
        "https://neuroconv.readthedocs.io/en/main/developer_guide/style_guide.html#style-guide-and-general-conventions%s",
        None,
    ),
    "black-coding-style": ("https://black.readthedocs.io/en/stable/the_black_code_style/current_style.html%s", None),
    "ncbi": ("https://www.ncbi.nlm.nih.gov/taxonomy%s", None),
    "ontobee": ("https://ontobee.org/%s", None),
}

# Use this for mapping for links to commonly used documentation
intersphinx_mapping = {
    "pynwb": ("https://pynwb.readthedocs.io/en/stable/", None),
    "nwb-schema": ("https://nwb-schema.readthedocs.io/en/latest/", None),
    "hdmf-schema": ("https://hdmf-common-schema.readthedocs.io/en/stable/", None),
}



"""Test the CLI calls related to streaming."""

import os
from pathlib import Path

import py
import py.path
import pytest

from nwbinspector.testing import check_streaming_tests_enabled

STREAMING_TESTS_ENABLED, DISABLED_STREAMING_TESTS_REASON = check_streaming_tests_enabled()
EXPECTED_REPORTS_FOLDER_PATH = Path(__file__).parent / "expected_reports"


@pytest.mark.skipif(not STREAMING_TESTS_ENABLED, reason=DISABLED_STREAMING_TESTS_REASON or "")
def test_dandiset_streaming_cli(tmpdir: py.path.local):
    tmpdir = Path(tmpdir)

    console_output_file_path = tmpdir / "test_console_output_dandiset_streaming_cli.txt"
    os.system(f"nwbinspector 000126 --stream > {console_output_file_path}")

    assert console_output_file_path.exists()

    with open(file=console_output_file_path, mode="r") as io:
        test_console_output = io.readlines()

    expected_output_file_path = EXPECTED_REPORTS_FOLDER_PATH / "000126_report.txt"
    with open(file=expected_output_file_path, mode="r") as io:
        expected_console_output = io.readlines()

    # Different platforms maybe have different indices for start and end of test reports
    report_start = test_console_output.index("0  CRITICAL\n")
    expected_report_length = 38
    report_end = report_start + expected_report_length
    assert test_console_output[report_start:report_end] == expected_console_output[14:]


@pytest.mark.skipif(not STREAMING_TESTS_ENABLED, reason=DISABLED_STREAMING_TESTS_REASON or "")
def test_dandiset_streaming_cli_with_version(tmpdir: py.path.local):
    tmpdir = Path(tmpdir)

    console_output_file_path = tmpdir / "test_console_output_dandiset_streaming_cli_with_version.txt"
    os.system(f"nwbinspector 000126 --version-id 0.210813.0327 --stream > {console_output_file_path}")

    assert console_output_file_path.exists()

    with open(file=console_output_file_path, mode="r") as io:
        test_console_output = io.readlines()

    expected_output_file_path = EXPECTED_REPORTS_FOLDER_PATH / "000126_report.txt"
    with open(file=expected_output_file_path, mode="r") as io:
        expected_console_output = io.readlines()

    # Different platforms maybe have different indices for start and end of test reports
    report_start = test_console_output.index("0  CRITICAL\n")
    expected_report_length = 38
    report_end = report_start + expected_report_length
    assert test_console_output[report_start:report_end] == expected_console_output[14:]


@pytest.mark.skipif(not STREAMING_TESTS_ENABLED, reason=DISABLED_STREAMING_TESTS_REASON or "")
def test_dandiset_streaming_cli_saved_report(tmpdir: py.path.local):
    tmpdir = Path(tmpdir)

    report_file_path = tmpdir / "test_dandiset_streaming_cli_saved_report.txt"
    console_output_file_path = tmpdir / "test_dandiset_streaming_cli_saved_report_console_output.txt"
    os.system(f"nwbinspector 000126 --stream --report-file-path {report_file_path} > {console_output_file_path}")

    assert report_file_path.exists()
    assert console_output_file_path.exists()

    with open(file=console_output_file_path, mode="r") as io:
        test_console_output = io.readlines()

    assert "Report saved to " in "".join(test_console_output[-3:])

    with open(file=report_file_path, mode="r") as io:
        test_report = io.readlines()

    expected_report_file_path = EXPECTED_REPORTS_FOLDER_PATH / "000126_report.txt"
    with open(file=expected_report_file_path, mode="r") as io:
        expected_report = io.readlines()

    # Different platforms maybe have different indices for start and end of test reports
    report_start = test_report.index("0  CRITICAL\n")
    expected_report_length = 38
    report_end = report_start + expected_report_length
    assert test_report[report_start:report_end] == expected_report[14:]


@pytest.mark.skipif(not STREAMING_TESTS_ENABLED, reason=DISABLED_STREAMING_TESTS_REASON or "")
def test_dandiset_streaming_cli_with_version_saved_report(tmpdir: py.path.local):
    tmpdir = Path(tmpdir)

    report_file_path = tmpdir / "test_dandiset_streaming_cli_with_version_saved_report.txt"
    console_output_file_path = tmpdir / "test_dandiset_streaming_cli_with_version_saved_report_console_output.txt"
    os.system(
        f"nwbinspector 000126 --version-id 0.210813.0327 --stream "
        f"--report-file-path {report_file_path} > {console_output_file_path}"
    )

    assert report_file_path.exists()
    assert console_output_file_path.exists()

    with open(file=console_output_file_path, mode="r") as io:
        test_console_output = io.readlines()

    assert "Report saved to " in "".join(test_console_output[-3:])

    with open(file=report_file_path, mode="r") as io:
        test_report = io.readlines()

    expected_report_file_path = EXPECTED_REPORTS_FOLDER_PATH / "000126_report.txt"
    with open(file=expected_report_file_path, mode="r") as io:
        expected_report = io.readlines()

    # Different platforms maybe have different indices for start and end of test reports
    report_start = test_report.index("0  CRITICAL\n")
    expected_report_length = 38
    report_end = report_start + expected_report_length
    assert test_report[report_start:report_end] == expected_report[14:]


@pytest.mark.skipif(not STREAMING_TESTS_ENABLED, reason=DISABLED_STREAMING_TESTS_REASON or "")
def test_inspect_url_cli(tmp_path):
    """Test that the CLI correctly handles inspecting a URL."""
    url = "https://dandiarchive.s3.amazonaws.com/blobs/11e/c89/11ec8933-1456-4942-922b-94e5878bb991"
    console_output_file = tmp_path / "url_console_output.txt"

    os.system(f"nwbinspector {url} " "--skip-validate " f"> {console_output_file}")

    # Read and check the contents of the console output file
    with open(console_output_file, "r") as file:
        output = file.read()
        assert "Subject species is missing." in output



import os
from datetime import datetime
from pathlib import Path
from shutil import rmtree
from tempfile import mkdtemp
from typing import Type, Union
from unittest import TestCase

import hdmf_zarr
import numpy as np
from hdmf.backends.io import HDMFIO
from hdmf.common import DynamicTable
from natsort import natsorted
from pynwb import NWBFile, TimeSeries
from pynwb.behavior import Position, SpatialSeries
from pynwb.file import Subject, TimeIntervals

from nwbinspector import (
    Importance,
    InspectorMessage,
    Severity,
    available_checks,
    inspect_all,
    inspect_nwbfile,
    inspect_nwbfile_object,
    load_config,
    register_check,
)
from nwbinspector.checks import (
    check_data_orientation,
    check_regular_timestamps,
    check_small_dataset_compression,
    check_subject_exists,
    check_timestamps_match_first_dimension,
)
from nwbinspector.testing import make_minimal_nwbfile
from nwbinspector.tools import BACKEND_IO_CLASSES
from nwbinspector.utils import FilePathType

IO_CLASSES_TO_BACKEND = {v: k for k, v in BACKEND_IO_CLASSES.items()}
EXPECTED_REPORTS_FOLDER_PATH = Path(__file__).parent / "expected_reports"


@register_check(importance=Importance.BEST_PRACTICE_VIOLATION, neurodata_type=DynamicTable)
def iterable_check_function(table: DynamicTable):
    for col in table.columns:
        yield InspectorMessage(message=f"Column: {col.name}")


def add_big_dataset_no_compression(nwbfile: NWBFile, zarr: bool = False) -> None:
    # Zarr automatically compresses by default
    # So to get a test case that is not compressed, forcibly disable the compressor
    if zarr:
        time_series = TimeSeries(
            name="test_time_series_1",
            data=hdmf_zarr.ZarrDataIO(np.zeros(shape=int(1.1e9 / np.dtype("float").itemsize)), compressor=False),
            rate=1.0,
            unit="",
        )
        nwbfile.add_acquisition(time_series)

        return

    time_series = TimeSeries(
        name="test_time_series_1", data=np.zeros(shape=int(1.1e9 / np.dtype("float").itemsize)), rate=1.0, unit=""
    )
    nwbfile.add_acquisition(time_series)


def add_regular_timestamps(nwbfile: NWBFile):
    regular_timestamps = np.arange(1.2, 11.2, 2)
    timestamps_length = len(regular_timestamps)
    time_series = TimeSeries(
        name="test_time_series_2",
        data=np.zeros(shape=(timestamps_length, timestamps_length - 1)),
        timestamps=regular_timestamps,
        unit="",
    )
    nwbfile.add_acquisition(time_series)


def add_flipped_data_orientation_to_acquisition(nwbfile: NWBFile):
    time_series = TimeSeries(name="my_spatial_series", data=np.zeros(shape=(2, 3)), unit="test_unit", rate=1.0)
    nwbfile.add_acquisition(time_series)


def add_flipped_data_orientation_to_processing(nwbfile: NWBFile):
    spatial_series = SpatialSeries(
        name="my_spatial_series", data=np.zeros(shape=(2, 3)), reference_frame="unknown", rate=1.0
    )
    module = nwbfile.create_processing_module(name="behavior", description="")
    module.add(Position(spatial_series=spatial_series))


def add_non_matching_timestamps_dimension(nwbfile: NWBFile):
    timestamps = [1.0, 2.1, 3.0]
    timestamps_length = len(timestamps)
    # Use __new__ and in_construct_mode=True to bypass the check in pynwb for data.shape[0] == len(timestamps)
    time_series = TimeSeries.__new__(TimeSeries, in_construct_mode=True)
    time_series.__init__(
        name="test_time_series_3",
        data=np.zeros(shape=(timestamps_length + 1, timestamps_length)),
        timestamps=timestamps,
        unit="",
    )
    nwbfile.add_acquisition(time_series)


def add_simple_table(nwbfile: NWBFile):
    time_intervals = TimeIntervals(name="test_table", description="desc")
    time_intervals.add_row(start_time=2.0, stop_time=3.0)
    time_intervals.add_row(start_time=1.0, stop_time=2.0)
    nwbfile.add_acquisition(time_intervals)


class TestInspectorOnBackend(TestCase):
    """A common helper class for testing the NWBInspector on files of a specific backend (HDF5 or Zarr)."""

    BackendIOClass: Type[HDMFIO]
    skip_validate = False  # TODO: can be removed once NWBZarrIO validation issues are resolved
    _backend_extensions = dict(zarr=".nwb.zarr", hdf5=".hdf5.nwb")

    @classmethod
    def get_extension(cls) -> str:
        backend_name = IO_CLASSES_TO_BACKEND[cls.BackendIOClass]
        return cls._backend_extensions[backend_name]

    @staticmethod
    def assertFileExists(path: Union[str, Path]):
        path = Path(path)
        assert path.exists()

    def assertLogFileContentsEqual(
        self,
        test_file_path: FilePathType,
        true_file_path: FilePathType,
        skip_first_newlines: bool = True,
        skip_last_newlines: bool = True,
    ):
        with open(file=test_file_path, mode="r") as test_file:
            test_file_lines = [x.rstrip("\n") for x in test_file.readlines()]
        with open(file=true_file_path, mode="r") as true_file:
            true_file_lines = [x.rstrip("\n") for x in true_file.readlines()]

        skip_first_n_lines = 0
        if skip_first_newlines:
            for line_number, test_line in enumerate(test_file_lines):
                if len(test_line) > 8:  # Can sometimes be a CLI specific byte such as '\x1b[0m\x1b[0m'
                    skip_first_n_lines = line_number
                    break

        skip_last_n_lines = 0
        if skip_last_newlines:
            for line_number, test_line in enumerate(test_file_lines[::-1]):
                if len(test_line) > 4:  # Can sometimes be a CLI specific byte such as '\x1b[0m'
                    skip_last_n_lines = line_number - 1  # Adjust for negative slicing
                    break

        for line_number, test_line in enumerate(test_file_lines):
            if "Timestamp: " in test_line:
                # Transform the test file header to match ground true example
                test_file_lines[line_number] = "Timestamp: 2022-04-01 13:32:13.756390-04:00"
                test_file_lines[line_number + 1] = "Platform: Windows-10-10.0.19043-SP0"
                test_file_lines[line_number + 2] = "NWBInspector version: 0.3.6"
            if ".nwb" in test_line:
                # Transform temporary testing path and formatted to hardcoded fake path
                suffix = self.get_extension()
                str_loc = test_line.find(suffix)
                correction_str = test_line.replace(test_line[5 : str_loc - 8], "./")  # noqa: E203 (black)
                test_file_lines[line_number] = correction_str
        self.assertEqual(first=test_file_lines[skip_first_n_lines : -(1 + skip_last_n_lines)], second=true_file_lines)


class TestInspectorAPIAndCLIHDF5(TestInspectorOnBackend):
    BackendIOClass = BACKEND_IO_CLASSES["hdf5"]
    skip_validate = False
    true_report_file_path = EXPECTED_REPORTS_FOLDER_PATH / "true_nwbinspector_default_report_hdf5.txt"
    maxDiff = None

    @classmethod
    def setUpClass(cls):
        cls.tempdir = Path(mkdtemp())
        cls.checks = [
            check_small_dataset_compression,
            check_regular_timestamps,
            check_data_orientation,
            check_timestamps_match_first_dimension,
        ]
        num_nwbfiles = 4
        nwbfiles = list()
        for j in range(num_nwbfiles):
            nwbfiles.append(make_minimal_nwbfile())
        add_big_dataset_no_compression(nwbfiles[0], zarr=cls.BackendIOClass is BACKEND_IO_CLASSES["zarr"])
        add_regular_timestamps(nwbfiles[0])
        add_flipped_data_orientation_to_processing(nwbfiles[0])
        add_non_matching_timestamps_dimension(nwbfiles[0])
        add_simple_table(nwbfiles[0])
        add_regular_timestamps(nwbfiles[1])
        # Third file to be left without violations
        add_non_matching_timestamps_dimension(nwbfiles[3])

        suffix = cls.get_extension()
        cls.nwbfile_paths = [str(cls.tempdir / f"testing{j}{suffix}") for j in range(num_nwbfiles)]
        cls.nwbfile_paths[3] = str(cls.tempdir / "._testing3.nwb")
        for nwbfile_path, nwbfile in zip(cls.nwbfile_paths, nwbfiles):
            with cls.BackendIOClass(path=nwbfile_path, mode="w") as io:
                io.write(nwbfile)

    @classmethod
    def tearDownClass(cls):
        rmtree(cls.tempdir, ignore_errors=True)

    def test_inspect_all(self):
        test_results = list(
            inspect_all(path=self.tempdir, select=[x.__name__ for x in self.checks], skip_validate=self.skip_validate)
        )
        true_results = [
            InspectorMessage(
                message="data is not compressed. Consider enabling compression when writing a dataset.",
                importance=Importance.BEST_PRACTICE_SUGGESTION,
                severity=Severity.LOW,
                check_function_name="check_small_dataset_compression",
                object_type="TimeSeries",
                object_name="test_time_series_1",
                location="/acquisition/test_time_series_1",
                file_path=self.nwbfile_paths[0],
            ),
            InspectorMessage(
                message=(
                    "TimeSeries appears to have a constant sampling rate. Consider specifying starting_time=1.2 "
                    "and rate=0.5 instead of timestamps."
                ),
                importance=Importance.BEST_PRACTICE_VIOLATION,
                severity=Severity.LOW,
                check_function_name="check_regular_timestamps",
                object_type="TimeSeries",
                object_name="test_time_series_2",
                location="/acquisition/test_time_series_2",
                file_path=self.nwbfile_paths[0],
            ),
            InspectorMessage(
                message=(
                    "Data may be in the wrong orientation. Time should be in the first dimension, and is usually "
                    "the longest dimension. Here, another dimension is longer."
                ),
                importance=Importance.CRITICAL,
                severity=Severity.LOW,
                check_function_name="check_data_orientation",
                object_type="SpatialSeries",
                object_name="my_spatial_series",
                location="/processing/behavior/Position/my_spatial_series",
                file_path=self.nwbfile_paths[0],
            ),
            InspectorMessage(
                message="The length of the first dimension of data (4) does not match the length of timestamps (3).",
                importance=Importance.CRITICAL,
                severity=Severity.LOW,
                check_function_name="check_timestamps_match_first_dimension",
                object_type="TimeSeries",
                object_name="test_time_series_3",
                location="/acquisition/test_time_series_3",
                file_path=self.nwbfile_paths[0],
            ),
            InspectorMessage(
                message=(
                    "TimeSeries appears to have a constant sampling rate. Consider specifying starting_time=1.2 "
                    "and rate=0.5 instead of timestamps."
                ),
                importance=Importance.BEST_PRACTICE_VIOLATION,
                severity=Severity.LOW,
                check_function_name="check_regular_timestamps",
                object_type="TimeSeries",
                object_name="test_time_series_2",
                location="/acquisition/test_time_series_2",
                file_path=self.nwbfile_paths[1],
            ),
        ]
        self.assertCountEqual(first=test_results, second=true_results)

    def test_inspect_all_parallel(self):
        test_results = list(
            inspect_all(
                path=Path(self.nwbfile_paths[0]).parent,
                select=[x.__name__ for x in self.checks],
                n_jobs=2,
                skip_validate=self.skip_validate,
            )
        )
        true_results = [
            InspectorMessage(
                message="data is not compressed. Consider enabling compression when writing a dataset.",
                importance=Importance.BEST_PRACTICE_SUGGESTION,
                severity=Severity.LOW,
                check_function_name="check_small_dataset_compression",
                object_type="TimeSeries",
                object_name="test_time_series_1",
                location="/acquisition/test_time_series_1",
                file_path=self.nwbfile_paths[0],
            ),
            InspectorMessage(
                message=(
                    "TimeSeries appears to have a constant sampling rate. Consider specifying starting_time=1.2 "
                    "and rate=0.5 instead of timestamps."
                ),
                importance=Importance.BEST_PRACTICE_VIOLATION,
                severity=Severity.LOW,
                check_function_name="check_regular_timestamps",
                object_type="TimeSeries",
                object_name="test_time_series_2",
                location="/acquisition/test_time_series_2",
                file_path=self.nwbfile_paths[0],
            ),
            InspectorMessage(
                message=(
                    "Data may be in the wrong orientation. Time should be in the first dimension, and is usually "
                    "the longest dimension. Here, another dimension is longer."
                ),
                importance=Importance.CRITICAL,
                severity=Severity.LOW,
                check_function_name="check_data_orientation",
                object_type="SpatialSeries",
                object_name="my_spatial_series",
                location="/processing/behavior/Position/my_spatial_series",
                file_path=self.nwbfile_paths[0],
            ),
            InspectorMessage(
                message=("The length of the first dimension of data (4) does not match the length of timestamps (3)."),
                importance=Importance.CRITICAL,
                severity=Severity.LOW,
                check_function_name="check_timestamps_match_first_dimension",
                object_type="TimeSeries",
                object_name="test_time_series_3",
                location="/acquisition/test_time_series_3",
                file_path=self.nwbfile_paths[0],
            ),
            InspectorMessage(
                message=(
                    "TimeSeries appears to have a constant sampling rate. Consider specifying starting_time=1.2 "
                    "and rate=0.5 instead of timestamps."
                ),
                importance=Importance.BEST_PRACTICE_VIOLATION,
                severity=Severity.LOW,
                check_function_name="check_regular_timestamps",
                object_type="TimeSeries",
                object_name="test_time_series_2",
                location="/acquisition/test_time_series_2",
                file_path=self.nwbfile_paths[1],
            ),
        ]
        self.assertCountEqual(first=test_results, second=true_results)

    def test_inspect_all_directory(self):
        """Test that inspect_all will find the file when given a valid path (in the case of Zarr, this path may be a directory)."""
        test_results = list(
            inspect_all(
                path=self.nwbfile_paths[0], select=[x.__name__ for x in self.checks], skip_validate=self.skip_validate
            )
        )
        self.assertGreater(len(test_results), 0)

    def test_inspect_nwbfile(self):
        test_results = list(
            inspect_nwbfile(nwbfile_path=self.nwbfile_paths[0], checks=self.checks, skip_validate=self.skip_validate)
        )
        true_results = [
            InspectorMessage(
                message="data is not compressed. Consider enabling compression when writing a dataset.",
                severity=Severity.LOW,
                importance=Importance.BEST_PRACTICE_SUGGESTION,
                check_function_name="check_small_dataset_compression",
                object_type="TimeSeries",
                object_name="test_time_series_1",
                location="/acquisition/test_time_series_1",
                file_path=self.nwbfile_paths[0],
            ),
            InspectorMessage(
                message=(
                    "TimeSeries appears to have a constant sampling rate. Consider specifying starting_time=1.2 and "
                    "rate=0.5 instead of timestamps."
                ),
                severity=Severity.LOW,
                importance=Importance.BEST_PRACTICE_VIOLATION,
                check_function_name="check_regular_timestamps",
                object_type="TimeSeries",
                object_name="test_time_series_2",
                location="/acquisition/test_time_series_2",
                file_path=self.nwbfile_paths[0],
            ),
            InspectorMessage(
                message=(
                    "Data may be in the wrong orientation. Time should be in the first dimension, and is usually the "
                    "longest dimension. Here, another dimension is longer."
                ),
                importance=Importance.CRITICAL,
                check_function_name="check_data_orientation",
                object_type="SpatialSeries",
                object_name="my_spatial_series",
                location="/processing/behavior/Position/my_spatial_series",
                file_path=self.nwbfile_paths[0],
            ),
            InspectorMessage(
                message="The length of the first dimension of data (4) does not match the length of timestamps (3).",
                importance=Importance.CRITICAL,
                check_function_name="check_timestamps_match_first_dimension",
                object_type="TimeSeries",
                object_name="test_time_series_3",
                location="/acquisition/test_time_series_3",
                file_path=self.nwbfile_paths[0],
            ),
        ]
        self.assertCountEqual(first=test_results, second=true_results)

    def test_inspect_nwbfile_importance_threshold_as_importance(self):
        test_results = list(
            inspect_nwbfile(
                nwbfile_path=self.nwbfile_paths[0],
                checks=self.checks,
                importance_threshold=Importance.CRITICAL,
                skip_validate=self.skip_validate,
            )
        )
        true_results = [
            InspectorMessage(
                message=(
                    "Data may be in the wrong orientation. Time should be in the first dimension, and is "
                    "usually the longest dimension. Here, another dimension is longer."
                ),
                importance=Importance.CRITICAL,
                check_function_name="check_data_orientation",
                object_type="SpatialSeries",
                object_name="my_spatial_series",
                location="/processing/behavior/Position/my_spatial_series",
                file_path=self.nwbfile_paths[0],
            ),
            InspectorMessage(
                message="The length of the first dimension of data (4) does not match the length of timestamps (3).",
                importance=Importance.CRITICAL,
                check_function_name="check_timestamps_match_first_dimension",
                object_type="TimeSeries",
                object_name="test_time_series_3",
                location="/acquisition/test_time_series_3",
                file_path=self.nwbfile_paths[0],
            ),
        ]
        self.assertCountEqual(first=test_results, second=true_results)

    def test_inspect_nwbfile_importance_threshold_as_string(self):
        test_results = list(
            inspect_nwbfile(
                nwbfile_path=self.nwbfile_paths[0],
                checks=self.checks,
                importance_threshold="CRITICAL",
                skip_validate=self.skip_validate,
            )
        )
        true_results = [
            InspectorMessage(
                message=(
                    "Data may be in the wrong orientation. Time should be in the first dimension, and is "
                    "usually the longest dimension. Here, another dimension is longer."
                ),
                importance=Importance.CRITICAL,
                check_function_name="check_data_orientation",
                object_type="SpatialSeries",
                object_name="my_spatial_series",
                location="/processing/behavior/Position/my_spatial_series",
                file_path=self.nwbfile_paths[0],
            ),
            InspectorMessage(
                message="The length of the first dimension of data (4) does not match the length of timestamps (3).",
                importance=Importance.CRITICAL,
                check_function_name="check_timestamps_match_first_dimension",
                object_type="TimeSeries",
                object_name="test_time_series_3",
                location="/acquisition/test_time_series_3",
                file_path=self.nwbfile_paths[0],
            ),
        ]
        self.assertCountEqual(first=test_results, second=true_results)

    def test_command_line_runs_cli_only(self):
        console_output_file = self.tempdir / "test_console_output.txt"
        os.system(
            f"nwbinspector {str(self.tempdir)} --overwrite --select check_timestamps_match_first_dimension,"
            "check_data_orientation,check_regular_timestamps,check_small_dataset_compression "
            "--modules random,math,datetime "
            "--skip-validate "
            f"> {console_output_file}"
        )
        self.assertLogFileContentsEqual(
            test_file_path=console_output_file,
            true_file_path=self.true_report_file_path,
            skip_first_newlines=True,
        )

    def test_command_line_runs_cli_only_parallel(self):
        console_output_file = self.tempdir / "test_console_output_2.txt"
        os.system(
            f"nwbinspector {str(self.tempdir)} --n-jobs 2 --overwrite --select check_timestamps_match_first_dimension,"
            "check_data_orientation,check_regular_timestamps,check_small_dataset_compression "
            "--skip-validate "
            f"> {console_output_file}"
        )
        self.assertLogFileContentsEqual(
            test_file_path=console_output_file,
            true_file_path=self.true_report_file_path,
            skip_first_newlines=True,
        )

    def test_command_line_saves_report(self):
        console_output_file = self.tempdir / "test_console_output_3.txt"
        os.system(
            f"nwbinspector {str(self.nwbfile_paths[0])} "
            f"--report-file-path {self.tempdir / 'test_nwbinspector_report_1.txt'} "
            "--skip-validate "
            f"> {console_output_file}"
        )
        self.assertFileExists(path=self.tempdir / "test_nwbinspector_report_1.txt")

    def test_command_line_organizes_levels(self):
        console_output_file = self.tempdir / "test_console_output_4.txt"
        os.system(
            f"nwbinspector {str(self.nwbfile_paths[0])} "
            f"--report-file-path {self.tempdir / 'test_nwbinspector_report_2.txt'} "
            "--levels importance,check_function_name,file_path "
            "--skip-validate "
            f"> {console_output_file}"
        )
        self.assertFileExists(path=self.tempdir / "test_nwbinspector_report_2.txt")

    def test_command_line_runs_saves_json(self):
        json_fpath = self.tempdir / "nwbinspector_results.json"
        os.system(f"nwbinspector {str(self.nwbfile_paths[0])} --json-file-path {json_fpath} " f"--skip-validate ")
        self.assertFileExists(path=json_fpath)

    def test_command_line_on_directory_matches_file(self):
        console_output_file = self.tempdir / "test_console_output_5.txt"
        os.system(
            f"nwbinspector {str(self.tempdir)} --overwrite --select check_timestamps_match_first_dimension,"
            "check_data_orientation,check_regular_timestamps,check_small_dataset_compression"
            f" --report-file-path {self.tempdir / 'test_nwbinspector_report_3.txt'} "
            "--skip-validate "
            f"> {console_output_file}"
        )
        self.assertLogFileContentsEqual(
            test_file_path=self.tempdir / "test_nwbinspector_report_3.txt",
            true_file_path=self.true_report_file_path,
            skip_first_newlines=True,
        )

    def test_iterable_check_function(self):
        test_results = list(
            inspect_nwbfile(
                nwbfile_path=self.nwbfile_paths[0], select=["iterable_check_function"], skip_validate=self.skip_validate
            )
        )
        true_results = [
            InspectorMessage(
                message="Column: start_time",
                importance=Importance.BEST_PRACTICE_VIOLATION,
                check_function_name="iterable_check_function",
                object_type="TimeIntervals",
                object_name="test_table",
                file_path=self.nwbfile_paths[0],
            ),
            InspectorMessage(
                message="Column: stop_time",
                importance=Importance.BEST_PRACTICE_VIOLATION,
                check_function_name="iterable_check_function",
                object_type="TimeIntervals",
                object_name="test_table",
                file_path=self.nwbfile_paths[0],
            ),
        ]
        self.assertCountEqual(first=test_results, second=true_results)

    def test_inspect_nwbfile_manual_iteration(self):
        generator = inspect_nwbfile(
            nwbfile_path=self.nwbfile_paths[0], checks=self.checks, skip_validate=self.skip_validate
        )
        message = next(generator)
        true_result = InspectorMessage(
            message="data is not compressed. Consider enabling compression when writing a dataset.",
            importance=Importance.BEST_PRACTICE_SUGGESTION,
            severity=Severity.LOW,
            check_function_name="check_small_dataset_compression",
            object_type="TimeSeries",
            object_name="test_time_series_1",
            location="/acquisition/test_time_series_1",
            file_path=self.nwbfile_paths[0],
        )
        self.assertEqual(message, true_result)

    def test_inspect_nwbfile_manual_iteration_stop(self):
        generator = inspect_nwbfile(
            nwbfile_path=self.nwbfile_paths[2], checks=self.checks, skip_validate=self.skip_validate
        )
        with self.assertRaises(expected_exception=StopIteration):
            next(generator)

    def test_inspect_nwbfile_dandi_config(self):
        config_checks = [check_subject_exists] + self.checks
        test_results = list(
            inspect_nwbfile(
                nwbfile_path=self.nwbfile_paths[0],
                checks=config_checks,
                config=load_config(filepath_or_keyword="dandi"),
                skip_validate=self.skip_validate,
            )
        )
        true_results = [
            InspectorMessage(
                message="Subject is missing.",
                importance=Importance.CRITICAL,  # Normally a BEST_PRACTICE_SUGGESTION
                check_function_name="check_subject_exists",
                object_type="NWBFile",
                object_name="root",
                location="/",
                file_path=self.nwbfile_paths[0],
            ),
            InspectorMessage(
                message=(
                    "Data may be in the wrong orientation. "
                    "Time should be in the first dimension, and is usually the longest dimension. "
                    "Here, another dimension is longer."
                ),
                importance=Importance.BEST_PRACTICE_VIOLATION,  # Normally CRITICAL, now a BEST_PRACTICE_VIOLATION
                check_function_name="check_data_orientation",
                object_type="SpatialSeries",
                object_name="my_spatial_series",
                location="/processing/behavior/Position/my_spatial_series",
                file_path=self.nwbfile_paths[0],
            ),
            InspectorMessage(
                message="data is not compressed. Consider enabling compression when writing a dataset.",
                importance=Importance.BEST_PRACTICE_SUGGESTION,
                check_function_name="check_small_dataset_compression",
                object_type="TimeSeries",
                object_name="test_time_series_1",
                location="/acquisition/test_time_series_1",
                file_path=self.nwbfile_paths[0],
            ),
            InspectorMessage(
                message=(
                    "TimeSeries appears to have a constant sampling rate. "
                    "Consider specifying starting_time=1.2 and rate=0.5 instead of timestamps."
                ),
                importance=Importance.BEST_PRACTICE_VIOLATION,
                check_function_name="check_regular_timestamps",
                object_type="TimeSeries",
                object_name="test_time_series_2",
                location="/acquisition/test_time_series_2",
                file_path=self.nwbfile_paths[0],
            ),
            InspectorMessage(
                message="The length of the first dimension of data (4) does not match the length of timestamps (3).",
                importance=Importance.CRITICAL,
                check_function_name="check_timestamps_match_first_dimension",
                object_type="TimeSeries",
                object_name="test_time_series_3",
                location="/acquisition/test_time_series_3",
                file_path=self.nwbfile_paths[0],
            ),
        ]
        self.assertCountEqual(first=test_results, second=true_results)


class TestInspectorAPIAndCLIZarr(TestInspectorAPIAndCLIHDF5):
    BackendIOClass = BACKEND_IO_CLASSES["zarr"]
    true_report_file_path = EXPECTED_REPORTS_FOLDER_PATH / "true_nwbinspector_default_report_zarr.txt"
    skip_validate = True


class TestDANDIConfigHDF5(TestInspectorOnBackend):
    BackendIOClass = BACKEND_IO_CLASSES["hdf5"]
    true_report_file_path = EXPECTED_REPORTS_FOLDER_PATH / "true_nwbinspector_report_with_dandi_config_hdf5.txt"
    skip_validate = False
    maxDiff = None

    @classmethod
    def setUpClass(cls):
        cls.tempdir = Path(mkdtemp())
        cls.checks = available_checks
        num_nwbfiles = 2
        nwbfiles = list()
        for j in range(num_nwbfiles):
            nwbfiles.append(make_minimal_nwbfile())
        add_big_dataset_no_compression(nwbfiles[0])
        add_regular_timestamps(nwbfiles[0])
        add_flipped_data_orientation_to_processing(nwbfiles[0])
        add_non_matching_timestamps_dimension(nwbfiles[0])
        add_simple_table(nwbfiles[0])
        add_flipped_data_orientation_to_acquisition(nwbfiles[1])

        suffix = cls.get_extension()
        cls.nwbfile_paths = [str(cls.tempdir / f"testing{j}{suffix}") for j in range(num_nwbfiles)]
        for nwbfile_path, nwbfile in zip(cls.nwbfile_paths, nwbfiles):
            with cls.BackendIOClass(path=nwbfile_path, mode="w") as io:
                io.write(nwbfile)

    @classmethod
    def tearDownClass(cls):
        rmtree(cls.tempdir, ignore_errors=True)

    def test_inspect_nwbfile_dandi_config_critical_only_entire_registry(self):
        test_results = list(
            inspect_nwbfile(
                nwbfile_path=self.nwbfile_paths[0],
                checks=available_checks,
                config=load_config(filepath_or_keyword="dandi"),
                importance_threshold=Importance.CRITICAL,
                skip_validate=self.skip_validate,
            )
        )
        true_results = [
            InspectorMessage(
                message="Subject is missing.",
                importance=Importance.CRITICAL,  # Normally a BEST_PRACTICE_SUGGESTION
                check_function_name="check_subject_exists",
                object_type="NWBFile",
                object_name="root",
                location="/",
                file_path=self.nwbfile_paths[0],
            ),
            InspectorMessage(
                message="The length of the first dimension of data (4) does not match the length of timestamps (3).",
                importance=Importance.CRITICAL,
                check_function_name="check_timestamps_match_first_dimension",
                object_type="TimeSeries",
                object_name="test_time_series_3",
                location="/acquisition/test_time_series_3",
                file_path=self.nwbfile_paths[0],
            ),
        ]
        self.assertCountEqual(first=test_results, second=true_results)

    def test_inspect_nwbfile_dandi_config_violation_and_above_entire_registry(self):
        test_results = list(
            inspect_nwbfile(
                nwbfile_path=self.nwbfile_paths[1],
                checks=available_checks,
                config=load_config(filepath_or_keyword="dandi"),
                importance_threshold=Importance.BEST_PRACTICE_VIOLATION,
                skip_validate=self.skip_validate,
            )
        )
        true_results = [
            InspectorMessage(
                message="Subject is missing.",
                importance=Importance.CRITICAL,  # Normally a BEST_PRACTICE_SUGGESTION
                check_function_name="check_subject_exists",
                object_type="NWBFile",
                object_name="root",
                location="/",
                file_path=self.nwbfile_paths[1],
            ),
            InspectorMessage(
                message=(
                    "Data may be in the wrong orientation. Time should be in the first dimension, "
                    "and is usually the longest dimension. Here, another dimension is longer."
                ),
                importance=Importance.BEST_PRACTICE_VIOLATION,
                severity=Severity.LOW,
                check_function_name="check_data_orientation",
                object_type="TimeSeries",
                object_name="my_spatial_series",
                location="/acquisition/my_spatial_series",
                file_path=self.nwbfile_paths[1],
            ),
        ]
        self.assertCountEqual(first=test_results, second=true_results)

    def test_inspect_nwbfile_dandi_config_critical_only_entire_registry_cli(self):
        console_output_file_path = self.tempdir / "test_console_output.txt"

        os.system(
            f"nwbinspector {str(self.tempdir)} --overwrite --config dandi --threshold BEST_PRACTICE_VIOLATION "
            f"--skip-validate "
            f"> {console_output_file_path}"
        )

        self.assertLogFileContentsEqual(
            test_file_path=console_output_file_path,
            true_file_path=self.true_report_file_path,
            skip_first_newlines=True,
        )


class TestDANDIConfigZarr(TestDANDIConfigHDF5):
    BackendIOClass = BACKEND_IO_CLASSES["zarr"]
    true_report_file_path = EXPECTED_REPORTS_FOLDER_PATH / "true_nwbinspector_report_with_dandi_config_zarr.txt"
    skip_validate = True


class TestCheckUniqueIdentifiersPassHDF5(TestInspectorOnBackend):
    BackendIOClass = BACKEND_IO_CLASSES["hdf5"]
    skip_validate = True
    maxDiff = None

    @classmethod
    def setUpClass(cls):
        cls.tempdir = Path(mkdtemp())
        num_nwbfiles = 3
        unique_id_nwbfiles = list()
        for j in range(num_nwbfiles):
            unique_id_nwbfiles.append(make_minimal_nwbfile())

        suffix = cls.get_extension()
        cls.unique_id_nwbfile_paths = [str(cls.tempdir / f"unique_id_testing{j}{suffix}") for j in range(num_nwbfiles)]
        for nwbfile_path, nwbfile in zip(cls.unique_id_nwbfile_paths, unique_id_nwbfiles):
            with cls.BackendIOClass(path=nwbfile_path, mode="w") as io:
                io.write(nwbfile)

    @classmethod
    def tearDownClass(cls):
        rmtree(cls.tempdir, ignore_errors=True)

    def test_check_unique_identifiers_pass(self):
        test_message = list(
            inspect_all(path=self.tempdir, select=["check_data_orientation"], skip_validate=self.skip_validate)
        )
        expected_message = []

        assert test_message == expected_message


class TestCheckUniqueIdentifiersFailHDF5(TestInspectorOnBackend):
    BackendIOClass = BACKEND_IO_CLASSES["hdf5"]
    skip_validate = True
    maxDiff = None

    @classmethod
    def setUpClass(cls):
        cls.tempdir = Path(mkdtemp())
        num_nwbfiles = 3
        non_unique_id_nwbfiles = list()
        for j in range(num_nwbfiles):
            non_unique_id_nwbfiles.append(
                NWBFile(
                    session_description="",
                    identifier="not a unique identifier!",
                    session_start_time=datetime.now().astimezone(),
                )
            )

        suffix = cls.get_extension()
        cls.non_unique_id_nwbfile_paths = [
            str(cls.tempdir / f"non_unique_id_testing{j}{suffix}") for j in range(num_nwbfiles)
        ]
        for nwbfile_path, nwbfile in zip(cls.non_unique_id_nwbfile_paths, non_unique_id_nwbfiles):
            with cls.BackendIOClass(path=nwbfile_path, mode="w") as io:
                io.write(nwbfile)

    @classmethod
    def tearDownClass(cls):
        rmtree(cls.tempdir, ignore_errors=True)

    def test_check_unique_identifiers_fail(self):
        test_messages = list(
            inspect_all(path=self.tempdir, select=["check_data_orientation"], skip_validate=self.skip_validate)
        )
        expected_messages = [
            InspectorMessage(
                message=(
                    "The identifier 'not a unique identifier!' is used across the .nwb files: "
                    f"{natsorted([Path(x).name for x in self.non_unique_id_nwbfile_paths])}. "
                    "The identifier of any NWBFile should be a completely unique value - "
                    "we recommend using uuid4 to achieve this."
                ),
                importance=Importance.CRITICAL,
                check_function_name="check_unique_identifiers",
                object_type="NWBFile",
                object_name="root",
                location="/",
                file_path=str(self.tempdir),
            )
        ]

        assert test_messages == expected_messages


class TestCheckUniqueIdentifiersPassZarr(TestCheckUniqueIdentifiersPassHDF5):
    BackendIOClass = BACKEND_IO_CLASSES["zarr"]


class TestCheckUniqueIdentifiersFailZarr(TestCheckUniqueIdentifiersFailHDF5):
    BackendIOClass = BACKEND_IO_CLASSES["zarr"]


def test_dandi_config_in_vitro_injection():
    """Test that a subject_id starting with 'protein' excludes meaningless CRITICAL-elevated subject checks."""
    nwbfile = make_minimal_nwbfile()
    nwbfile.subject = Subject(
        subject_id="proteinCaMPARI3", description="A detailed description about the in vitro setup."
    )
    config = load_config(filepath_or_keyword="dandi")
    importance_threshold = "CRITICAL"
    messages = list(
        inspect_nwbfile_object(nwbfile_object=nwbfile, config=config, importance_threshold=importance_threshold)
    )
    assert messages == []


def test_dandi_config_in_vitro_injection_safe():
    """Test the safe subject ID retrieval of the in vitro injection."""
    nwbfile = make_minimal_nwbfile()
    nwbfile.subject = Subject(subject_id=None, description="A detailed description about the in vitro setup.")
    config = load_config(filepath_or_keyword="dandi")
    messages = list(inspect_nwbfile_object(nwbfile_object=nwbfile, config=config))
    assert len(messages) != 0



"""Test the API functions related to streaming."""

import pytest

from nwbinspector import (
    Importance,
    InspectorMessage,
    inspect_all,
    inspect_dandi_file_path,
    inspect_dandiset,
    inspect_url,
)
from nwbinspector.testing import check_streaming_tests_enabled

STREAMING_TESTS_ENABLED, DISABLED_STREAMING_TESTS_REASON = check_streaming_tests_enabled()


@pytest.mark.skipif(not STREAMING_TESTS_ENABLED, reason=DISABLED_STREAMING_TESTS_REASON or "")
def test_inspect_all_streaming():
    dandiset_id = "000126"
    select = ["check_subject_species_exists"]

    test_messages = list(inspect_all(path=dandiset_id, stream=True, select=select))
    assert len(test_messages) == 1

    test_message = test_messages[0]
    expected_message = InspectorMessage(
        message="Subject species is missing.",
        importance=Importance.CRITICAL,  # Uses dandi config by default
        check_function_name="check_subject_species_exists",
        object_type="Subject",
        object_name="subject",
        location="/general/subject",
        file_path="sub-1/sub-1.nwb",
    )

    assert test_message == expected_message


@pytest.mark.skipif(not STREAMING_TESTS_ENABLED, reason=DISABLED_STREAMING_TESTS_REASON or "")
def test_inspect_dandiset():
    dandiset_id = "000126"
    select = ["check_subject_species_exists"]

    test_messages = list(inspect_dandiset(dandiset_id=dandiset_id, select=select))
    assert len(test_messages) == 1

    test_message = test_messages[0]
    expected_message = InspectorMessage(
        message="Subject species is missing.",
        importance=Importance.CRITICAL,  # Uses dandi config by default
        check_function_name="check_subject_species_exists",
        object_type="Subject",
        object_name="subject",
        location="/general/subject",
        file_path="sub-1/sub-1.nwb",
    )

    assert test_message == expected_message


@pytest.mark.skipif(not STREAMING_TESTS_ENABLED, reason=DISABLED_STREAMING_TESTS_REASON or "")
def test_inspect_dandi_file_path():
    dandiset_id = "000126"
    dandi_file_path = "sub-1/sub-1.nwb"
    select = ["check_subject_species_exists"]

    test_messages = list(
        inspect_dandi_file_path(dandi_file_path=dandi_file_path, dandiset_id=dandiset_id, select=select)
    )
    assert len(test_messages) == 1

    test_message = test_messages[0]
    expected_message = InspectorMessage(
        message="Subject species is missing.",
        importance=Importance.CRITICAL,  # Uses dandi config by default
        check_function_name="check_subject_species_exists",
        object_type="Subject",
        object_name="subject",
        location="/general/subject",
        file_path=dandi_file_path,
    )

    assert test_message == expected_message


@pytest.mark.skipif(not STREAMING_TESTS_ENABLED, reason=DISABLED_STREAMING_TESTS_REASON or "")
def test_inspect_url():
    url = "https://dandiarchive.s3.amazonaws.com/blobs/11e/c89/11ec8933-1456-4942-922b-94e5878bb991"
    select = ["check_subject_species_exists"]

    test_messages = list(inspect_url(url=url, select=select))
    assert len(test_messages) == 1

    test_message = test_messages[0]
    expected_message = InspectorMessage(
        message="Subject species is missing.",
        importance=Importance.CRITICAL,  # Uses dandi config by default
        check_function_name="check_subject_species_exists",
        object_type="Subject",
        object_name="subject",
        location="/general/subject",
        file_path=url,
    )

    assert test_message == expected_message



"""All tests that specifically require streaming to be enabled (i.e., ROS3 version of h5py, fsspec, etc.)."""

import pytest

from nwbinspector.testing import check_hdf5_io_open, check_streaming_tests_enabled
from nwbinspector.tools import read_nwbfile

STREAMING_TESTS_ENABLED, DISABLED_STREAMING_TESTS_REASON = check_streaming_tests_enabled()
PERSISTENT_READ_NWBFILE_HDF5_EXAMPLE_HTTPS = (
    "https://dandi-api-staging-dandisets.s3.amazonaws.com/blobs/80d/80f/80d80f55-f8a1-4318-b17e-ce55f4dd2620"
)
PERSISTENT_READ_NWBFILE_HDF5_EXAMPLE_S3 = (
    "s3://dandi-api-staging-dandisets/blobs/80d/80f/80d80f55-f8a1-4318-b17e-ce55f4dd2620"
)
PERSISTENT_READ_NWBFILE_ZARR_EXAMPLE_HTTPS = (
    "https://dandi-api-staging-dandisets.s3.amazonaws.com/zarr/63f06140-8d0c-4db4-81cc-812ed4e4db03"
)
PERSISTENT_READ_NWBFILE_ZARR_EXAMPLE_S3 = "s3://dandi-api-staging-dandisets/zarr/63f06140-8d0c-4db4-81cc-812ed4e4db03"


# These will move to PyNWB when the time is right
@pytest.mark.skipif(not STREAMING_TESTS_ENABLED, reason=DISABLED_STREAMING_TESTS_REASON or "")
def test_hdf5_fsspec_https():
    nwbfile = read_nwbfile(
        nwbfile_path=PERSISTENT_READ_NWBFILE_HDF5_EXAMPLE_HTTPS,
        backend="hdf5",  # TODO: cannot current auto-detect backend when streaming
        method="fsspec",
    )
    assert check_hdf5_io_open(io=nwbfile.read_io)

    nwbfile.read_io.close()
    assert not check_hdf5_io_open(io=nwbfile.read_io)


@pytest.mark.skipif(not STREAMING_TESTS_ENABLED, reason=DISABLED_STREAMING_TESTS_REASON or "")
def test_hdf5_fsspec_s3():
    nwbfile = read_nwbfile(
        nwbfile_path=PERSISTENT_READ_NWBFILE_HDF5_EXAMPLE_S3,
        backend="hdf5",  # TODO: cannot current auto-detect backend when streaming
        method="fsspec",
    )
    assert check_hdf5_io_open(io=nwbfile.read_io)

    nwbfile.read_io.close()
    assert not check_hdf5_io_open(io=nwbfile.read_io)


@pytest.mark.skipif(not STREAMING_TESTS_ENABLED, reason=DISABLED_STREAMING_TESTS_REASON or "")
def test_hdf5_ros3_https():
    nwbfile = read_nwbfile(
        nwbfile_path=PERSISTENT_READ_NWBFILE_HDF5_EXAMPLE_HTTPS,
        backend="hdf5",  # TODO: cannot current auto-detect backend when streaming
        method="ros3",
    )
    assert check_hdf5_io_open(io=nwbfile.read_io)

    nwbfile.read_io.close()
    assert not check_hdf5_io_open(io=nwbfile.read_io)


# Zarr files not working yet with streaming

# @pytest.mark.skipif(not STREAMING_TESTS_ENABLED, reason=DISABLED_STREAMING_TESTS_REASON or "")
# def test_zarr_fsspec_streaming_https():
#     nwbfile = read_nwbfile(
#         nwbfile_path=PERSISTENT_READ_NWBFILE_ZARR_EXAMPLE_HTTPS,
#         method="fsspec",
#     )
#     assert check_zarr_io_open(io=nwbfile.read_io)

#     nwbfile.read_io.close()
#     assert not check_zarr_io_open(io=nwbfile.read_io)


# @pytest.mark.skipif(not STREAMING_TESTS_ENABLED, reason=DISABLED_STREAMING_TESTS_REASON or "")
# def test_zarr_fsspec_streaming_s3():
#     nwbfile = read_nwbfile(
#         nwbfile_path=PERSISTENT_READ_NWBFILE_ZARR_EXAMPLE_S3,
#         method="fsspec",
#     )
#     assert check_zarr_io_open(io=nwbfile.read_io)

#     nwbfile.read_io.close()
#     assert not check_zarr_io_open(io=nwbfile.read_io)


# @pytest.mark.skipif(not STREAMING_TESTS_ENABLED, reason=DISABLED_STREAMING_TESTS_REASON or "")
# def test_zarr_ros3_error():
#     test that an error is raised with ros3 driver on zarr (not applicable)
